# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"UniqueNamer": "01_putils.ipynb",
         "FunctionsList": "01_putils.ipynb",
         "get_drive": "01_putils.ipynb",
         "stringIntListToListOfInts": "01_putils.ipynb",
         "stringListToListOfStrings": "01_putils.ipynb",
         "listNumsToString": "01_putils.ipynb",
         "sigmoid": "01_putils.ipynb",
         "is_in_notebooks": "01_putils.ipynb",
         "BaseFunction": "02_functions.ipynb",
         "Subtract": "02_functions.ipynb",
         "Proportional": "02_functions.ipynb",
         "Variable": "02_functions.ipynb",
         "PassOn": "02_functions.ipynb",
         "GreaterThan": "02_functions.ipynb",
         "Constant": "02_functions.ipynb",
         "Step": "02_functions.ipynb",
         "Integration": "02_functions.ipynb",
         "IntegrationDual": "02_functions.ipynb",
         "Sigmoid": "02_functions.ipynb",
         "WeightedSum": "02_functions.ipynb",
         "IndexedParameter": "02_functions.ipynb",
         "PCTNode": "03_nodes.ipynb",
         "PCTNodeData": "03_nodes.ipynb",
         "FunctionsData": "04_hierarchy.ipynb",
         "PCTHierarchy": "04_hierarchy.ipynb",
         "OpenAIGym": "05_environments.ipynb",
         "CartPoleV1": "05_environments.ipynb",
         "PendulumV0": "05_environments.ipynb",
         "VelocityModel": "05_environments.ipynb",
         "DummyModel": "05_environments.ipynb",
         "EnvironmentFactory": "05_environments.ipynb",
         "BaseArchitecture": "06_architectures.ipynb",
         "ProportionalArchitecture": "06_architectures.ipynb",
         "LevelKey": "06_architectures.ipynb",
         "ControlUnitIndices": "06_architectures.ipynb",
         "DynamicArchitecture": "06_architectures.ipynb",
         "BaseErrorType": "07_errors.ipynb",
         "RootSumSquaredError": "07_errors.ipynb",
         "RootMeanSquareError": "07_errors.ipynb",
         "BaseErrorCollector": "07_errors.ipynb",
         "TotalError": "07_errors.ipynb",
         "InputsError": "07_errors.ipynb",
         "ErrorFactory": "07_errors.ipynb"}

modules = ["putils.py",
           "functions.py",
           "nodes.py",
           "hierarchy.py",
           "environments.py",
           "architectures.py",
           "errors.py"]

doc_url = "https://perceptualrobots.github.io/pct/"

git_url = "https://github.com/perceptualrobots/pct/tree/master/"

def custom_doc_links(name): return None
